/****************************************************************

  Traction Software, Inc. Confidential and Proprietary Information

  Copyright (c) 1996-2015 Traction Software, Inc.
  All rights reserved.

****************************************************************/

// PLEASE DO NOT DELETE THIS LINE -- make copyright depends on it.
package com.traction.view.listener;

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.Date;

import javax.mail.internet.MimeUtility;

import tsi.admin.ServerAdmin;

import com.traction.sdk.*;
import com.traction.sdk.event.Event;
import com.traction.sdk.form.StandardDateRange;
import com.traction.sdk.mail.AddressException;
import com.traction.sdk.mail.MailConfigurationException;
import com.traction.sdk.mail.Message;
import com.traction.sdk.props.GetProperty;
import com.traction.sdk.props.NativeTypeConversion;
import com.traction.sdk.props.PropNamespace;
import com.traction.sdk.props.SimpleProperties;
import com.traction.sdk.sdl.CompiledTemplate;
import com.traction.sdk.sdl.Scope;
import com.traction.sdk.sdl.SdlEvaluator;
import com.traction.sdk.sdl.SdlFactory;
import com.traction.sdk.sdl.Tag;
import com.traction.sdk.sdl.tag.SectionsTag;
import com.traction.sdk.util.MailUtil;
import com.traction.sdk.util.Provider;
import com.traction.sdk.util.SectionsUtil;
import com.traction.sdk.util.log.Debug;
import com.traction.sdk.view.Section;
import com.traction.settings.Select;
import com.traction.view.listener.notifier.transport.EmailTransport;

public class EmailNotifier2 extends EmailNotifier {

    public static final String PROJECT_LIST_PROP_NAME = "emailnotifier2_projects";
    public static final String DEFAULT_SDL_TEMPLATE = "emailnotifier2_sdl";
    public static final String PROJECT_LIST_SUBSCRIBED_PROP_NAME = PropNamespace.ns(PROJECT_LIST_PROP_NAME,
										    Select.MultiSelectEditor2.STORE_UNUSED);

    public static final String EVENTS_SUBSCRIBED_PROP_NAME = "emailnotifier2_events";
    public static final String USE_EVENTS_PROP_NAME        = "emailnotifier2_use_events";

    public static final String SUBSCRIPTION_SECTIONS_PROP_NAME     = "emailnotifier2_sections";
    public static final String SUBSCRIPTION_USE_SECTIONS_PROP_NAME = "emailnotifier2_use_sections";

    public static final String EMAIL_NOTIFICATION2_SELF_PREFERENCE_NAME = "emailnotifier2_enabled_own";
    public static final String EMAIL_NOTIFICATION2_OTHERS_PREFERENCE_NAME = "emailnotifier2_enabled_others";
    public static final String EMAIL_NOTIFICATION2_EMAIL_ADDRESS_PREFERENCE_NAME = "emailnotifier2_address";

    public static final String LISTENER_CONFIG_PROP_NAME_MESSAGE_BODY_SDL_TEMPLATE = "sdltemplate_setting";

    @Override
    protected String getEmailAddress() {
	UserPreferences up = context.getUser().getUserPreferences();
	return SimpleProperties.loadString(up.getProperties(context),
					   EMAIL_NOTIFICATION2_EMAIL_ADDRESS_PREFERENCE_NAME,
					   up.getPreferredEmail());
    }

    @Override
    public String getRunAsUser() {
	return context.getUser().getName();
    }

    /**
     * Tests whether notification is allowed to proceed, according to various
     * rules.  If the listener configuration file is misconfigured; or if its
     * configuration does not allow the current user to receive notification; or
     * if the  listener's configuration specifies that events in the given
     * project are to be ignored; or if the given user has not asked to receive
     * notification, notification should be skipped.
     * @return false if any of the conditions above holds; true otherwise.
     * @throws AuthenticationException if the entry from the context (and/or, in
     * in the case of an edit or reclassify event, the entry to which that entry
     * refers) is not readable by the user that the listener is running as.
     * @throws JournalException if the entry from the context
     * (and/or, in the case of an edit or reclassify event, the entry to which
     * that entry refers) is not readable by the user that the listener is
     * running as.
     */
    @Override
    protected boolean skipNotification()
	throws AuthenticationException, JournalException {

        if (Notifier.skipEmailRecordNoContent(context)) {
            return true;
        }
        if (Notifier.skipReclassifyRecordNoVisibleLabels(context)) {
            return true;
        }

        // If _either_ of the following holds for the user who is supposed to
	// receive notification, then the notification should be skipped:
	// 1) the user is the author of this entry (i.e., the user triggered
	// this listener) but the user has opted out of receiving notifications
	// about events that he himself triggered; 2) the user isn't the author
	// of this entry (i.e., some other user triggered this listener) but
	// the user has opted out of receiving notifications about events that
	// other users triggered.
	User user = context.getUser();
	String username = user.getName();
	GetProperty prefs = user.getUserPreferences().getProperties(context);
	boolean notifyauthorself = SimpleProperties.loadBoolean(prefs, EMAIL_NOTIFICATION2_SELF_PREFERENCE_NAME, false);
	boolean notifyauthorothers = SimpleProperties.loadBoolean(prefs, EMAIL_NOTIFICATION2_OTHERS_PREFERENCE_NAME, false);

	// So that we don't have to make a request to the journal to retrieve
	// the entry (in order to get the author) right now, we should do this
	// check up front:  if the user has opted out of notification whether or
	// not the user is the author, we can stop right here.  This may speed
	// things up when skipping notification.  [shep 02.Sep.2005]
	if (!notifyauthorself && !notifyauthorothers) {
	    Debug.event.println("[", getListenerName(), "] Skipping email notification for user \"", username,
				"\" because that user has opted out of receiving notifications",
				" for events (triggered by self or others).");
	    return true;
	}

	TractionId id = context.getUrl().getEntry();

	// If we try to get the entry and can't, it means that the listener is
	// mis-configured because the user property is missing altogether, has
	// no value, or specifies an unrecognized user or a user who does not
	// have permission to read the entry; or that the id is out of range
	// (the entry does not exist).
	entry = context.getJournal().getEntry(id, jreq);

	boolean isauthor = entry.isMine();
	if (isauthor) {
	    if (!notifyauthorself) {
		Debug.event.println("[", getListenerName(), "] Skipping email notification for user \"", username,
				    "\" because that user has opted out of receiving notifications",
				    " for events he/she triggered.");
		return true; // user doesn't want notification of an entry he/she authored
	    }
	}
	else {
	    if (!notifyauthorothers) {
		Debug.event.println("[", getListenerName(), "] Skipping email notification for user \"", username,
				    "\" because that user has opted out of receiving notifications",
				    " for events other users triggered.");
		return true; // user doesn't want notification of an entry another user authored
	    }
	}

	// See if I care about this event (or if I'm even managing my
	// subscriptions via event type filtering).
	if (SimpleProperties.loadBoolean(prefs, USE_EVENTS_PROP_NAME, false)) {
	    ArrayList<String> events = new ArrayList<String>();
	    SimpleProperties.loadList(prefs, EVENTS_SUBSCRIBED_PROP_NAME, events, SimpleProperties.LIST_SINGLE_PROPERTY);
	    if (!events.contains(event)) {
		if (Debug.event.enabled()) {
		    Debug.event.println("[", getListenerName(), "] Skipping email notification for user \"", username,
					"\" because that user has opted out of receiving notifications",
					" for ", event, " events.");
		}
		return true;
	    }
	}

	// If the entry is an edit event, use the entry that is being
	// updated to create the body; else use the given entry itself.  If this
	// fails with an exception, it means that the entry id of the updated
	// entry is out of bounds or more likely that the entry being updated
	// cannot be read by the user that the listener is running as.
	if (event.equals(Event.EDIT)) {
	    scopeEntry = context.getJournal().getEntry(entry.getEntryUpdated(), jreq);
	} else {
	    scopeEntry = entry;
	}
	Debug.event.println("[", getListenerName(), "] Using ",
			    scopeEntry.getTractionId().toString(),
			    " as the entry for the message body of notification for ", event,
			    " event, entry ", entry.getTractionId().toString(), ".");

	// Make sure the user is subscribed to this project.
	Project proj = id.getProject();
	String projidStr = String.valueOf(proj.getId());
	// not_subscribed_list is a flag that indicates whether or not
	// the project list is the list of projects to which the user
	// is not subscribed, in which case the user is subscribed to
	// all other accessible projects (rather than being a list of
	// those to which the user is subscribed, in which case the
	// list will be an exhaustive list of projects to which the
	// user is subscribed).
	boolean not_subscribed_list = SimpleProperties.loadBoolean(prefs, PROJECT_LIST_SUBSCRIBED_PROP_NAME, true);
	ArrayList<String> proj_list = new ArrayList<String>();
	SimpleProperties.loadList(prefs, PROJECT_LIST_PROP_NAME, proj_list, SimpleProperties.LIST_SINGLE_PROPERTY);
	// if list doesn't contain project, and list represents
	// subscribed list, then we don't notify; and if list contains
	// project, and list represents not-subscribed list, then we
	// don't notify.
	if (proj_list.contains(projidStr) == not_subscribed_list) {
	    Debug.event.println("[", getListenerName(), "] Skipping ", event,
				" notification for entry ", id.toString(),
				" because the user is not subscribed to email notifications for the ",
				proj.getName(), " project.");
	    return true;
	}

	if (SimpleProperties.loadBoolean(prefs, SUBSCRIPTION_USE_SECTIONS_PROP_NAME, true) &&
	    !containsHitAroundEntry(context, entry, "user", SUBSCRIPTION_SECTIONS_PROP_NAME)) {
	    Debug.event.println("[", getListenerName(), "] Skipping ", event,
				" notification for entry ", context.getUrl().getEntry().toString(),
				" because ", context.getUser().getName(),
				" does not have a notification query that would include that entry).");
	    return true;
	}

	// If all tests succeeded, return false:  do not skip the notification.
	Debug.event.println("[", getListenerName(), "] Notification will proceed (event=", event,
			    "; entry=", context.getUrl().getEntry().toString(),
			    "; recipient=\"", username, "\").");
	return false;

    }

    @Override
    protected void setupMessage()
	throws ConfigurationException, IOException, UnsupportedEncodingException {

	String sdltemplate = getMessageBodyTemplate();

	if (sdltemplate == null) {
	    throw new ConfigurationException("[" + getListenerName() + "] Configuration Error: can't determine which SDL template to use.");
	}

	Debug.event.println("[", getListenerName(), "] Using SDL template ", sdltemplate);

	// Try to create the template, and make sure creation succeeded.
	SdlFactory sdl = context.getSdlFactory();
	CompiledTemplate template = sdl.createTemplate(sdltemplate);
	if (template == null) {
	    throw new ConfigurationException("Configuration Error: Configuration Error: the SDL template " +
					     sdltemplate +
					     " cannot be found.");
	}

	// Prepare to evaluate the SDL template for the message body
	// for now, just use the "html" version.  Future will use
	// "html" or "text" depending upon setting.
	Tag bodyBlock = template.getBlock(sdl, "html");

	// SDL used for tokens assume that the entry whose token is
	// being evaluated is in scope by default. Since this method
	// sometimes asks for an Entry's content directly, we will for
	// now set the current entry on the context's default scope.
	// This is fairly safe in this limited case because we know
	// that event listener views are centered on a single entry
	// only.  Later, when we fix the underlying problem at the
	// renderer level, we will change this back to setting the
	// scopeentry on the created scope here, as commented out
	// below.  See AKJ6983.06. [shep 27.Mar.2008]
	context.getCurrentScope().setCurrentEntry(scopeEntry);
	Scope scope = sdl.createScope(template, context, bodyBlock);
	scope.setCurrentEntry(scopeEntry);

	scope.setGlobalVariable(EVENT_TYPE_SDL_VARIABLE_NAME, event);

	if (event.equals(Event.EDIT)) {
	    modifyScopeForEdit(scope);
	}
	if (event.equals(Event.RECLASSIFY)) {
	    modifyScopeForReclassify(scope);
	}

	// Try to create the message.
	try {
	    message = context.getSdkFactory().createSmtpMessage(context);
	}
	catch (AuthenticationException ae) {
	    throw new ConfigurationException("Configuration Error: The user \"" +
					     SimpleProperties.loadString(config, RUN_AS_USER_PROPERTY, null) +
					     "\" being used to run this listener does not have permission to send mail.");
	}
	catch (MailConfigurationException mce) {
	    throw new ConfigurationException("Configuration Error: " + mce.getMessage());
	}

	TractionId id = context.getUrl().getEntry();
	String sender = getSenderAddress(id, scopeEntry);
	String friendlySenderName = getSenderFriendlyName(id, scopeEntry);

	if (friendlySenderName != null) {
	    // Encode the friendly name. See AKJ3058. [shep 27.Jul.2005]
	    sender = MailUtil.makeFriendlyNameAddress(sender, MimeUtility.encodeText(friendlySenderName, "utf-8", null));
	}

	try {
	    message.addFrom(sender);
	}
	catch (AddressException ae) {
	    throw new ConfigurationException("Configuration Error: Sender email address (\"" +
					     ae.getAddress() + "\") is invalid.\nOriginal Message:  " +
					     ae.getMessage());
	}

	JournalProperties jProps = context.getJournal().getJournalProperties();
	String replyTo = jProps.getMailboxAddress();
	if (replyTo != null && jProps.getMailboxEnabled()) {
	    replyTo = EmailTransport.getModifiedReplyToForTargetId(replyTo, event, scopeEntry, context);
	    try {
		String friendlyReplyTo = jProps.getMailboxFriendlyName();
		if (friendlyReplyTo != null) {
		    replyTo = MailUtil.makeFriendlyNameAddress(replyTo, MimeUtility.encodeText(friendlyReplyTo, "utf-8", null));
		}
		message.addReplyTo(replyTo);
	    } catch (AddressException ae) {
		throw new ConfigurationException("Configuration Error: Server mailbox address (\"" +
						 ae.getAddress() + "\") is invalid.\nOriginal Message:  " +
						 ae.getMessage());
	    }
	}

	// Retrieve the user's preferred email address from the user's
	// preferences.  If the user does not have a preferrred email address,
	// or if the address is invalid, throw a ConfigurationException.  If the
	// address is valid, use it as the recipient address for this message
	// (To: field).
	//
	// Include the user's full name in the recipient address, as the
	// friendly name.  And encode it in a mail-safe fashion.
	UserPreferences up = context.getUser().getUserPreferences();
	recipient = SimpleProperties.loadString(up.getProperties(context),
						EMAIL_NOTIFICATION2_EMAIL_ADDRESS_PREFERENCE_NAME,
						up.getPreferredEmail());
	if (recipient == null || recipient.trim().equals("")) {
	    throw new ConfigurationException("Configuration Error: Cannot notify "
					     + context.getUser().getName() + " of " + event +
					     " event (" + context.getUrl().getEntry().toString() +
					     ") because this user has no preferred email address.");
	} else {
	    String fullname = up.getFullName();
	    if (fullname != null && !fullname.trim().equals("")) {
		try {
		    recipient = MailUtil.makeFriendlyNameAddress(recipient, MimeUtility.encodeText(fullname, "utf-8", null));
		} catch (UnsupportedEncodingException uee) {
		    uee.printStackTrace(Debug.log);
		}
	    }
	}

	try {
	    message.addRecipient(recipient, Message.ADDRTYPE_TO);
	} catch (AddressException ae) {
	    throw new ConfigurationException("Configuration Error: Preferred email address for user " +
					     context.getUser().getName() + " (\"" +
					     ae.getAddress() + "\") is invalid.\nOriginal Message:  " +
					     ae.getMessage());
	}

	if (SimpleProperties.loadBoolean(context.getUrl(), "test", false)) {
	    ArrayList<String> testRecipients = new ArrayList<String>();
	    SimpleProperties.loadList(context.getViewAssembly().getConfiguration(), "test_recipients", testRecipients, SimpleProperties.LIST_SINGLE_PROPERTY);
	    for (String testBcc : testRecipients) {
		try {
		    message.addRecipient(testBcc, Message.ADDRTYPE_BCC);
		    Debug.event.println("Adding BCC ", testBcc);
		} catch (AddressException ae) {
		    throw new ConfigurationException("Configuration Error: Test recipient " +
						     testBcc + " (\"" +
						     ae.getAddress() + "\") is invalid.\nOriginal Message:  " +
						     ae.getMessage());
		}
	    }
	}

	handleSubjectAndBody(scope, bodyBlock);

    }

    @Override
    protected final String getListenerName() {
	return context.getViewAssembly().getConfiguration().getName();
    }

    protected String getMessageBodyTemplate() {
	// The name of the server setting property that stores the
	// spec for the SDL template to use for the message body.
	String templatePropName = SimpleProperties.loadString(config,
	                                                      LISTENER_CONFIG_PROP_NAME_MESSAGE_BODY_SDL_TEMPLATE,
	                                                      PropNamespace.ns(getListenerName(),
	                                                                       "sdl"));
	return SimpleProperties.loadString(context.getServer().getSettings(),
					   templatePropName,
					   DEFAULT_SDL_TEMPLATE);
    }

    protected String getSenderAddress(TractionId id, Entry scopeentry) throws ConfigurationException {
	return
	    testForDefault(EmailTransport.getServerConfiguredSenderAddress(context,
									   event,
									   EmailTransport.getTargetAuthor(context, id, event, scopeentry),
									   getDefaultSenderAddress()));
    }

    protected String getDefaultSenderAddress() {
	return MailUtil.getRawAddressFromFriendlyEncoding(ServerAdmin.getServerDataInterpreter().getDigestSender());
    }

    protected String getSenderFriendlyName(TractionId id, Entry scopeentry) {
	return
	    EmailTransport.getSenderFriendlyName(context,
						 EmailTransport.getTargetAuthor(context, id, event, scopeentry),
						 entry.getTractionId().getProject());
    }

    protected final String testForDefault(String address) throws ConfigurationException {
	if (address == null || "".equals(address.trim())) {
	    throw new ConfigurationException("[" + getListenerName() + "] Configuration Error: unable to use '" + address + "' as From address.");
	}
	return address;
    }

    private static final String TEST_HIT_SDL_TEMPLATE_SPEC = "com.traction.sdl.admin.settings#sectionlist.hashit";

    /**
     * Returns true if the sections whose data is defined in the given
     * type of store under the given property name contain at least
     * one hit in a time slice from 1ms before to 1ms after the given
     * entry was posted.
     */
    public static final boolean containsHitAroundEntry(Context context, Entry entry, String sectionStoreType, String sectionPropName) {

	SdlEvaluator sdl = context.getSdlFactory().createSdlEvaluator(context, TEST_HIT_SDL_TEMPLATE_SPEC);
	if (!sdl.isValid()) {
	    Debug.logWarning("Unable to find SDL template '" + TEST_HIT_SDL_TEMPLATE_SPEC + "'.");
	    return false;
	}

	CJournalRequest origJreq = context.getDefaultJournalRequest();
	Scope scope = sdl.getScope();
	scope.setCurrentEntry(entry);
	scope.setGlobalVariable(SectionsTag.VAR_SECTIONS,
	                        SectionsUtil.getSectionsFromStore(context,
	                                                          Section.PropertyType.Subscription,
	                                                          sectionStoreType,
	                                                          sectionPropName,
	                                                          getProjectProvider(origJreq),
	                                                          getUserProvider(context)));
	scope.setGlobalVariable("hit", "false");

	JournalRequest jreq = origJreq.copy();
	long entryDate = entry.getDate().getTime();
	Date sDate = new Date(entryDate - 1L);
	Date eDate = new Date(entryDate + 1L);
	jreq.setFocusRange(new StandardDateRange(sDate, eDate));

	scope.setCurrentJournalRequest(jreq);
	context.getViewAssembly().getView().setMainJournalRequest(jreq);
	context.setDefaultJournalRequest(jreq);

	try {
	    sdl.eval(context.getSdkFactory().createNullPrintWriter());
	    Object var = scope.getGlobalVariable("hit");
	    if (var == null) {
	        return false;
	    }
	    if (var instanceof Boolean) {
	        return ((Boolean) var).booleanValue();
	    }
	    return NativeTypeConversion.stringToBoolean(var.toString(), false);
	}
	catch (IOException | RuntimeException e) {
	    Debug.logWarning(Debug.event, e, "Problem evaluating the SDL template '" + TEST_HIT_SDL_TEMPLATE_SPEC + "'");
	    return false;
	}
	finally {
	    jreq = origJreq.copy();
	    context.setDefaultJournalRequest(jreq);
	    context.getViewAssembly().getView().setMainJournalRequest(jreq);
	}

    }

    private static final Provider<Void,Project> getProjectProvider(final CJournalRequest jrequest) {
        return new Provider<Void,Project>() {
            @Override
            public final Project getValue(Void data) {
                return jrequest.getOnlyProject();
            }
        };
    }

    private static final Provider<Void,User> getUserProvider(final Context context) {
        return new Provider<Void,User>() {
            @Override
            public final User getValue(Void data) {
                return context.getUser();
            }
        };
    }

}
